@@ -3,6 +3,7 @@ package main
 import (
 	"bufio"
 	"context"
+	"encoding/json"
 	"fmt"
 	"io"
 	"log"
@@ -326,21 +327,16 @@ func servectlcmd(ctx context.Context, ctl *ctl, shutdown func()) {
 				log.Check(err, "closing temporary message file")
 			}
 		}()
-		mw := &message.Writer{Writer: msgFile}
+		mw := message.NewWriter(msgFile)
 		ctl.xwriteok()
 
 		ctl.xstreamto(mw)
 		err = msgFile.Sync()
 		ctl.xcheck(err, "syncing message to storage")
-		msgPrefix := []byte{}
-		if !mw.HaveHeaders {
-			msgPrefix = []byte("\r\n\r\n")
-		}
 
 		m := &store.Message{
-			Received:  time.Now(),
-			Size:      int64(len(msgPrefix)) + mw.Size,
-			MsgPrefix: msgPrefix,
+			Received: time.Now(),
+			Size:     mw.Size,
 		}
 
 		a.WithWLock(func() {
@@ -619,6 +615,12 @@ func servectlcmd(ctx context.Context, ctl *ctl, shutdown func()) {
 		account := ctl.xread()
 		acc, err := store.OpenAccount(account)
 		ctl.xcheck(err, "open account")
+		defer func() {
+			if acc != nil {
+				err := acc.Close()
+				log.Check(err, "closing account after retraining")
+			}
+		}()
 
 		acc.WithWLock(func() {
 			conf, _ := acc.Conf()
@@ -649,6 +651,7 @@ func servectlcmd(ctx context.Context, ctl *ctl, shutdown func()) {
 			// Read through messages with junk or nonjunk flag set, and train them.
 			var total, trained int
 			q := bstore.QueryDB[store.Message](ctx, acc.DB)
+			q.FilterEqual("Expunged", false)
 			err = q.ForEach(func(m store.Message) error {
 				total++
 				ok, err := acc.TrainMessage(ctx, ctl.log, jf, m)
@@ -665,9 +668,270 @@ func servectlcmd(ctx context.Context, ctl *ctl, shutdown func()) {
 			jf = nil
 			ctl.xcheck(err, "closing junk filter")
 		})
-
 		ctl.xwriteok()
 
+	case "recalculatemailboxcounts":
+		/* protocol:
+		> "recalculatemailboxcounts"
+		> account
+		< "ok" or error
+		< stream
+		*/
+		account := ctl.xread()
+		acc, err := store.OpenAccount(account)
+		ctl.xcheck(err, "open account")
+		defer func() {
+			if acc != nil {
+				err := acc.Close()
+				log.Check(err, "closing account after recalculating mailbox counts")
+			}
+		}()
+		ctl.xwriteok()
+
+		w := ctl.writer()
+
+		acc.WithWLock(func() {
+			var changes []store.Change
+			err = acc.DB.Write(ctx, func(tx *bstore.Tx) error {
+				return bstore.QueryTx[store.Mailbox](tx).ForEach(func(mb store.Mailbox) error {
+					mc, err := mb.CalculateCounts(tx)
+					if err != nil {
+						return fmt.Errorf("calculating counts for mailbox %q: %w", mb.Name, err)
+					}
+
+					if !mb.HaveCounts || mc != mb.MailboxCounts {
+						_, err := fmt.Fprintf(w, "for %s setting new counts %s (was %s)\n", mb.Name, mc, mb.MailboxCounts)
+						ctl.xcheck(err, "write")
+						mb.HaveCounts = true
+						mb.MailboxCounts = mc
+						if err := tx.Update(&mb); err != nil {
+							return fmt.Errorf("storing new counts for %q: %v", mb.Name, err)
+						}
+						changes = append(changes, mb.ChangeCounts())
+					}
+					return nil
+				})
+			})
+			ctl.xcheck(err, "write transaction for mailbox counts")
+
+			store.BroadcastChanges(acc, changes)
+		})
+		w.xclose()
+
+	case "fixmsgsize":
+		/* protocol:
+		> "fixmsgsize"
+		> account or empty
+		< "ok" or error
+		< stream
+		*/
+
+		accountOpt := ctl.xread()
+		ctl.xwriteok()
+		w := ctl.writer()
+
+		var foundProblem bool
+		const batchSize = 10000
+
+		xfixmsgsize := func(accName string) {
+			acc, err := store.OpenAccount(accName)
+			ctl.xcheck(err, "open account")
+			defer func() {
+				err := acc.Close()
+				log.Check(err, "closing account after fixing message sizes")
+			}()
+
+			total := 0
+			var lastID int64
+			for {
+				var n int
+
+				acc.WithRLock(func() {
+					mailboxCounts := map[int64]store.Mailbox{} // For broadcasting.
+
+					// Don't process all message in one transaction, we could block the account for too long.
+					err := acc.DB.Write(ctx, func(tx *bstore.Tx) error {
+						q := bstore.QueryTx[store.Message](tx)
+						q.FilterEqual("Expunged", false)
+						q.FilterGreater("ID", lastID)
+						q.Limit(batchSize)
+						q.SortAsc("ID")
+						return q.ForEach(func(m store.Message) error {
+							lastID = m.ID
+
+							p := acc.MessagePath(m.ID)
+							st, err := os.Stat(p)
+							if err != nil {
+								mb := store.Mailbox{ID: m.MailboxID}
+								if xerr := tx.Get(&mb); xerr != nil {
+									_, werr := fmt.Fprintf(w, "get mailbox id %d for message with file error: %v\n", mb.ID, xerr)
+									ctl.xcheck(werr, "write")
+								}
+								_, werr := fmt.Fprintf(w, "checking file %s for message %d in mailbox %q (id %d): %v (continuing)\n", p, m.ID, mb.Name, mb.ID, err)
+								ctl.xcheck(werr, "write")
+								return nil
+							}
+							filesize := st.Size()
+							correctSize := int64(len(m.MsgPrefix)) + filesize
+							if m.Size == correctSize {
+								return nil
+							}
+
+							foundProblem = true
+
+							mb := store.Mailbox{ID: m.MailboxID}
+							if err := tx.Get(&mb); err != nil {
+								_, werr := fmt.Fprintf(w, "get mailbox id %d for message with file size mismatch: %v\n", mb.ID, err)
+								ctl.xcheck(werr, "write")
+							}
+							_, err = fmt.Fprintf(w, "fixing message %d in mailbox %q (id %d) with incorrect size %d, should be %d (len msg prefix %d + on-disk file %s size %d)\n", m.ID, mb.Name, mb.ID, m.Size, correctSize, len(m.MsgPrefix), p, filesize)
+							ctl.xcheck(err, "write")
+
+							// We assume that the original message size was accounted as stored in the mailbox
+							// total size. If this isn't correct, the user can always run
+							// recalculatemailboxcounts.
+							mb.Size -= m.Size
+							mb.Size += correctSize
+							if err := tx.Update(&mb); err != nil {
+								return fmt.Errorf("update mailbox counts: %v", err)
+							}
+							mailboxCounts[mb.ID] = mb
+
+							m.Size = correctSize
+
+							mr := acc.MessageReader(m)
+							part, err := message.EnsurePart(log, false, mr, m.Size)
+							if err != nil {
+								_, werr := fmt.Fprintf(w, "parsing message %d again: %v (continuing)\n", m.ID, err)
+								ctl.xcheck(werr, "write")
+							}
+							m.ParsedBuf, err = json.Marshal(part)
+							if err != nil {
+								return fmt.Errorf("marshal parsed message: %v", err)
+							}
+							total++
+							n++
+							if err := tx.Update(&m); err != nil {
+								return fmt.Errorf("update message: %v", err)
+							}
+							return nil
+						})
+
+					})
+					ctl.xcheck(err, "find and fix wrong message sizes")
+
+					var changes []store.Change
+					for _, mb := range mailboxCounts {
+						changes = append(changes, mb.ChangeCounts())
+					}
+					store.BroadcastChanges(acc, changes)
+				})
+				if n < batchSize {
+					break
+				}
+			}
+			_, err = fmt.Fprintf(w, "%d message size(s) fixed for account %s\n", total, accName)
+			ctl.xcheck(err, "write")
+		}
+
+		if accountOpt != "" {
+			xfixmsgsize(accountOpt)
+		} else {
+			for i, accName := range mox.Conf.Accounts() {
+				var line string
+				if i > 0 {
+					line = "\n"
+				}
+				_, err := fmt.Fprintf(w, "%sFixing message sizes in account %s...\n", line, accName)
+				ctl.xcheck(err, "write")
+				xfixmsgsize(accName)
+			}
+		}
+		if foundProblem {
+			_, err := fmt.Fprintf(w, "\nProblems were found and fixed. You should invalidate messages stored at imap clients with the \"mox bumpuidvalidity account [mailbox]\" command.\n")
+			ctl.xcheck(err, "write")
+		}
+
+		w.xclose()
+
+	case "reparse":
+		/* protocol:
+		> "reparse"
+		> account or empty
+		< "ok" or error
+		< stream
+		*/
+
+		accountOpt := ctl.xread()
+		ctl.xwriteok()
+		w := ctl.writer()
+
+		const batchSize = 100
+
+		xreparseAccount := func(accName string) {
+			acc, err := store.OpenAccount(accName)
+			ctl.xcheck(err, "open account")
+			defer func() {
+				err := acc.Close()
+				log.Check(err, "closing account after reparsing messages")
+			}()
+
+			total := 0
+			var lastID int64
+			for {
+				var n int
+				// Don't process all message in one transaction, we could block the account for too long.
+				err := acc.DB.Write(ctx, func(tx *bstore.Tx) error {
+					q := bstore.QueryTx[store.Message](tx)
+					q.FilterEqual("Expunged", false)
+					q.FilterGreater("ID", lastID)
+					q.Limit(batchSize)
+					q.SortAsc("ID")
+					return q.ForEach(func(m store.Message) error {
+						lastID = m.ID
+						mr := acc.MessageReader(m)
+						p, err := message.EnsurePart(log, false, mr, m.Size)
+						if err != nil {
+							_, err := fmt.Fprintf(w, "parsing message %d: %v (continuing)\n", m.ID, err)
+							ctl.xcheck(err, "write")
+						}
+						m.ParsedBuf, err = json.Marshal(p)
+						if err != nil {
+							return fmt.Errorf("marshal parsed message: %v", err)
+						}
+						total++
+						n++
+						if err := tx.Update(&m); err != nil {
+							return fmt.Errorf("update message: %v", err)
+						}
+						return nil
+					})
+
+				})
+				ctl.xcheck(err, "update messages with parsed mime structure")
+				if n < batchSize {
+					break
+				}
+			}
+			_, err = fmt.Fprintf(w, "%d message(s) reparsed for account %s\n", total, accName)
+			ctl.xcheck(err, "write")
+		}
+
+		if accountOpt != "" {
+			xreparseAccount(accountOpt)
+		} else {
+			for i, accName := range mox.Conf.Accounts() {
+				var line string
+				if i > 0 {
+					line = "\n"
+				}
+				_, err := fmt.Fprintf(w, "%sReparsing account %s...\n", line, accName)
+				ctl.xcheck(err, "write")
+				xreparseAccount(accName)
+			}
+		}
+		w.xclose()
+
 	case "backup":
 		backupctl(ctx, ctl)
 
