@@ -21,12 +21,9 @@ import (
 
 	"github.com/mjl-/mox/dmarcrpt"
 	"github.com/mjl-/mox/dns"
-	"github.com/mjl-/mox/mlog"
 	"github.com/mjl-/mox/mox-"
 )
 
-var xlog = mlog.New("dmarcdb")
-
 var (
 	DBTypes = []any{DomainFeedback{}} // Types stored in DB.
 	DB      *bstore.DB                // Exported for backups.
