@@ -18,6 +18,13 @@ import (
 	"github.com/mjl-/mox/mox-"
 )
 
+func tcheck(t *testing.T, err error, msg string) {
+	t.Helper()
+	if err != nil {
+		t.Fatalf("%s: %s", msg, err)
+	}
+}
+
 func TestWebserver(t *testing.T) {
 	os.RemoveAll("../testdata/webserver/data")
 	mox.ConfigStaticPath = "../testdata/webserver/mox.conf"
