@@ -33,7 +33,6 @@ import (
 	"github.com/mjl-/mox/dmarcrpt"
 	"github.com/mjl-/mox/dns"
 	"github.com/mjl-/mox/dnsbl"
-	"github.com/mjl-/mox/http"
 	"github.com/mjl-/mox/message"
 	"github.com/mjl-/mox/mlog"
 	"github.com/mjl-/mox/mox-"
@@ -45,6 +44,7 @@ import (
 	"github.com/mjl-/mox/tlsrpt"
 	"github.com/mjl-/mox/tlsrptdb"
 	"github.com/mjl-/mox/updates"
+	"github.com/mjl-/mox/webadmin"
 )
 
 var (
@@ -132,6 +132,15 @@ var commands = []struct {
 	{"tlsrpt parsereportmsg", cmdTLSRPTParsereportmsg},
 	{"version", cmdVersion},
 
+	{"bumpuidvalidity", cmdBumpUIDValidity},
+	{"reassignuids", cmdReassignUIDs},
+	{"fixuidmeta", cmdFixUIDMeta},
+	{"fixmsgsize", cmdFixmsgsize},
+	{"reparse", cmdReparse},
+	{"ensureparsed", cmdEnsureParsed},
+	{"recalculatemailboxcounts", cmdRecalculateMailboxCounts},
+	{"message parse", cmdMessageParse},
+
 	// Not listed.
 	{"helpall", cmdHelpall},
 	{"junk analyze", cmdJunkAnalyze},
@@ -139,12 +148,7 @@ var commands = []struct {
 	{"junk play", cmdJunkPlay},
 	{"junk test", cmdJunkTest},
 	{"junk train", cmdJunkTrain},
-	{"bumpuidvalidity", cmdBumpUIDValidity},
-	{"reassignuids", cmdReassignUIDs},
-	{"fixuidmeta", cmdFixUIDMeta},
 	{"dmarcdb addreport", cmdDMARCDBAddReport},
-	{"ensureparsed", cmdEnsureParsed},
-	{"message parse", cmdMessageParse},
 	{"tlsrptdb addreport", cmdTLSRPTDBAddReport},
 	{"updates addsigned", cmdUpdatesAddSigned},
 	{"updates genkey", cmdUpdatesGenkey},
@@ -154,6 +158,7 @@ var commands = []struct {
 	{"gentestdata", cmdGentestdata},
 	{"ximport maildir", cmdXImportMaildir},
 	{"ximport mbox", cmdXImportMbox},
+	{"openaccounts", cmdOpenaccounts},
 }
 
 var cmds []cmd
@@ -376,6 +381,11 @@ func mustLoadConfig() {
 }
 
 func main() {
+	// CheckConsistencyOnClose is true by default, for all the test packages. A regular
+	// mox server should never use it. But integration tests enable it again with a
+	// flag.
+	store.CheckConsistencyOnClose = false
+
 	log.SetFlags(0)
 
 	// If invoked as sendmail, e.g. /usr/sbin/sendmail, we do enough so cron can get a
@@ -392,6 +402,7 @@ func main() {
 	flag.StringVar(&mox.ConfigStaticPath, "config", envString("MOXCONF", "config/mox.conf"), "configuration file, other config files are looked up in the same directory, defaults to $MOXCONF with a fallback to mox.conf")
 	flag.StringVar(&loglevel, "loglevel", "", "if non-empty, this log level is set early in startup")
 	flag.BoolVar(&pedantic, "pedantic", false, "protocol violations result in errors instead of accepting/working around them")
+	flag.BoolVar(&store.CheckConsistencyOnClose, "checkconsistency", false, "dangerous option for testing only, enables data checks that abort/panic when inconsistencies are found")
 
 	var cpuprofile, memprofile string
 	flag.StringVar(&cpuprofile, "cpuprof", "", "store cpu profile to file")
@@ -502,7 +513,9 @@ are printed.
 		c.Usage()
 	}
 
-	_, errs := mox.ParseConfig(context.Background(), mox.ConfigStaticPath, true, false, false)
+	mox.FilesImmediate = true
+
+	_, errs := mox.ParseConfig(context.Background(), mox.ConfigStaticPath, true, true, false)
 	if len(errs) > 1 {
 		log.Printf("multiple errors:")
 		for _, err := range errs {
@@ -777,7 +790,7 @@ func cmdConfigDNSCheck(c *cmd) {
 		log.Fatalf("%s", err)
 	}()
 
-	printResult := func(name string, r http.Result) {
+	printResult := func(name string, r webadmin.Result) {
 		if len(r.Errors) == 0 && len(r.Warnings) == 0 {
 			return
 		}
@@ -790,7 +803,7 @@ func cmdConfigDNSCheck(c *cmd) {
 		}
 	}
 
-	result := http.Admin{}.CheckDomain(context.Background(), args[0])
+	result := webadmin.Admin{}.CheckDomain(context.Background(), args[0])
 	printResult("IPRev", result.IPRev.Result)
 	printResult("MX", result.MX.Result)
 	printResult("TLS", result.TLS.Result)
@@ -1431,11 +1444,13 @@ headers prepended.
 		c.Usage()
 	}
 
+	clog := mlog.New("dkimsign")
+
 	msgf, err := os.Open(args[0])
 	xcheckf(err, "open message")
 	defer msgf.Close()
 
-	p, err := message.Parse(msgf)
+	p, err := message.Parse(clog, true, msgf)
 	xcheckf(err, "parsing message")
 
 	if len(p.Envelope.From) != 1 {
@@ -1579,7 +1594,7 @@ can be found in message headers.
 
 	data, err := io.ReadAll(os.Stdin)
 	xcheckf(err, "read message")
-	dmarcFrom, _, err := message.From(bytes.NewReader(data))
+	dmarcFrom, _, err := message.From(mlog.New("dmarcverify"), false, bytes.NewReader(data))
 	xcheckf(err, "extract dmarc from message")
 
 	const ignoreTestMode = false
@@ -1609,10 +1624,12 @@ understand email deliverability problems.
 		c.Usage()
 	}
 
+	clog := mlog.New("dmarcparsereportmsg")
+
 	for _, arg := range args {
 		f, err := os.Open(arg)
 		xcheckf(err, "open %q", arg)
-		feedback, err := dmarcrpt.ParseMessageReport(f)
+		feedback, err := dmarcrpt.ParseMessageReport(clog, f)
 		xcheckf(err, "parse report in %q", arg)
 		meta := feedback.ReportMetadata
 		fmt.Printf("Report: period %s-%s, organisation %q, reportID %q, %s\n", time.Unix(meta.DateRange.Begin, 0).UTC().String(), time.Unix(meta.DateRange.End, 0).UTC().String(), meta.OrgName, meta.ReportID, meta.Email)
@@ -1661,9 +1678,11 @@ func cmdDMARCDBAddReport(c *cmd) {
 
 	mustLoadConfig()
 
+	clog := mlog.New("dmarcdbaddreport")
+
 	fromdomain := xparseDomain(args[0], "domain")
 	fmt.Fprintln(os.Stderr, "reading report message from stdin")
-	report, err := dmarcrpt.ParseMessageReport(os.Stdin)
+	report, err := dmarcrpt.ParseMessageReport(clog, os.Stdin)
 	xcheckf(err, "parse message")
 	err = dmarcdb.AddReport(context.Background(), report, fromdomain)
 	xcheckf(err, "add dmarc report")
@@ -1700,10 +1719,12 @@ The report is printed in formatted JSON.
 		c.Usage()
 	}
 
+	clog := mlog.New("tlsrptparsereportmsg")
+
 	for _, arg := range args {
 		f, err := os.Open(arg)
 		xcheckf(err, "open %q", arg)
-		report, err := tlsrpt.ParseMessage(f)
+		report, err := tlsrpt.ParseMessage(clog, f)
 		xcheckf(err, "parse report in %q", arg)
 		// todo future: only print the highlights?
 		enc := json.NewEncoder(os.Stdout)
@@ -1843,11 +1864,13 @@ func cmdTLSRPTDBAddReport(c *cmd) {
 
 	mustLoadConfig()
 
+	clog := mlog.New("tlsrptdbaddreport")
+
 	// First read message, to get the From-header. Then parse it as TLSRPT.
 	fmt.Fprintln(os.Stderr, "reading report message from stdin")
 	buf, err := io.ReadAll(os.Stdin)
 	xcheckf(err, "reading message")
-	part, err := message.Parse(bytes.NewReader(buf))
+	part, err := message.Parse(clog, true, bytes.NewReader(buf))
 	xcheckf(err, "parsing message")
 	if part.Envelope == nil || len(part.Envelope.From) != 1 {
 		log.Fatalf("message must have one From-header")
@@ -1855,7 +1878,7 @@ func cmdTLSRPTDBAddReport(c *cmd) {
 	from := part.Envelope.From[0]
 	domain := xparseDomain(from.Host, "domain")
 
-	report, err := tlsrpt.ParseMessage(bytes.NewReader(buf))
+	report, err := tlsrpt.ParseMessage(clog, bytes.NewReader(buf))
 	xcheckf(err, "parsing tls report in message")
 
 	mailfrom := from.User + "@" + from.Host // todo future: should escape and such
@@ -1980,85 +2003,15 @@ func cmdVersion(c *cmd) {
 	fmt.Println(moxvar.Version)
 }
 
-func cmdEnsureParsed(c *cmd) {
-	c.unlisted = true
-	c.params = "account"
-	c.help = "Ensure messages in the database have a ParsedBuf."
-	var all bool
-	c.flag.BoolVar(&all, "all", false, "store new parsed message for all messages")
-	args := c.Parse()
-	if len(args) != 1 {
-		c.Usage()
-	}
-
-	mustLoadConfig()
-	a, err := store.OpenAccount(args[0])
-	xcheckf(err, "open account")
-	defer func() {
-		if err := a.Close(); err != nil {
-			log.Printf("closing account: %v", err)
-		}
-	}()
-
-	n := 0
-	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
-		q := bstore.QueryTx[store.Message](tx)
-		q.FilterFn(func(m store.Message) bool {
-			return all || m.ParsedBuf == nil
-		})
-		l, err := q.List()
-		if err != nil {
-			return fmt.Errorf("list messages: %v", err)
-		}
-		for _, m := range l {
-			mr := a.MessageReader(m)
-			p, err := message.EnsurePart(mr, m.Size)
-			if err != nil {
-				log.Printf("parsing message %d: %v (continuing)", m.ID, err)
-			}
-			m.ParsedBuf, err = json.Marshal(p)
-			if err != nil {
-				return fmt.Errorf("marshal parsed message: %v", err)
-			}
-			if err := tx.Update(&m); err != nil {
-				return fmt.Errorf("update message: %v", err)
-			}
-			n++
-		}
-		return nil
-	})
-	xcheckf(err, "update messages with parsed mime structure")
-	fmt.Printf("%d messages updated\n", n)
-}
-
-func cmdMessageParse(c *cmd) {
-	c.unlisted = true
-	c.params = "message.eml"
-	c.help = "Parse message, print JSON representation."
-
-	args := c.Parse()
-	if len(args) != 1 {
-		c.Usage()
-	}
-
-	f, err := os.Open(args[0])
-	xcheckf(err, "open")
-	defer f.Close()
-
-	part, err := message.Parse(f)
-	xcheckf(err, "parsing message")
-	err = part.Walk(nil)
-	xcheckf(err, "parsing nested parts")
-	enc := json.NewEncoder(os.Stdout)
-	enc.SetIndent("", "\t")
-	err = enc.Encode(part)
-	xcheckf(err, "write")
-}
-
 func cmdBumpUIDValidity(c *cmd) {
-	c.unlisted = true
 	c.params = "account [mailbox]"
-	c.help = "Change the IMAP UID validity of the mailbox, causing IMAP clients to refetch messages."
+	c.help = `Change the IMAP UID validity of the mailbox, causing IMAP clients to refetch messages.
+
+This can be useful after manually repairing metadata about the account/mailbox.
+
+Opens account database file directly. Ensure mox does not have the account
++open, or is not running.
+`
 	args := c.Parse()
 	if len(args) != 1 && len(args) != 2 {
 		c.Usage()
@@ -2101,7 +2054,6 @@ func cmdBumpUIDValidity(c *cmd) {
 }
 
 func cmdReassignUIDs(c *cmd) {
-	c.unlisted = true
 	c.params = "account [mailboxid]"
 	c.help = `Reassign UIDs in one mailbox or all mailboxes in an account and bump UID validity, causing IMAP clients to refetch messages.
 
@@ -2135,17 +2087,22 @@ open, or is not running.
 	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
 		// Reassign UIDs, going per mailbox. We assign starting at 1, only changing the
 		// message if it isn't already at the intended UID. Doing it in this order ensures
-		// we don't get into trouble with duplicate UIDs for a mailbox.
+		// we don't get into trouble with duplicate UIDs for a mailbox. We assign a new
+		// modseq. Not strictly needed, for doesn't hurt.
+		modseq, err := a.NextModSeq(tx)
+		xcheckf(err, "assigning next modseq")
+
 		q := bstore.QueryTx[store.Message](tx)
 		if len(args) == 2 {
 			q.FilterNonzero(store.Message{MailboxID: mailboxID})
 		}
 		q.SortAsc("MailboxID", "UID")
-		err := q.ForEach(func(m store.Message) error {
+		err = q.ForEach(func(m store.Message) error {
 			uidlasts[m.MailboxID]++
 			uid := uidlasts[m.MailboxID]
 			if m.UID != uid {
 				m.UID = uid
+				m.ModSeq = modseq
 				if err := tx.Update(&m); err != nil {
 					return fmt.Errorf("updating uid for message: %v", err)
 				}
@@ -2190,7 +2147,6 @@ open, or is not running.
 }
 
 func cmdFixUIDMeta(c *cmd) {
-	c.unlisted = true
 	c.params = "account"
 	c.help = `Fix inconsistent UIDVALIDITY and UIDNEXT in messages/mailboxes/account.
 
@@ -2234,9 +2190,10 @@ open, or is not running.
 			} else if err != nil {
 				return fmt.Errorf("finding message with max uid in mailbox: %w", err)
 			}
+			olduidnext := mb.UIDNext
 			mb.UIDNext = m.UID + 1
+			log.Printf("fixing uidnext to %d (max uid is %d, old uidnext was %d) for mailbox %q (id %d)", mb.UIDNext, m.UID, olduidnext, mb.Name, mb.ID)
 			if err := tx.Update(&mb); err != nil {
-				log.Printf("fixing uidnext to %d (max uid is %d) for mailbox id %d", mb.UIDNext, m.UID, mb.ID)
 				return fmt.Errorf("updating mailbox uidnext: %v", err)
 			}
 			return nil
@@ -2261,3 +2218,188 @@ open, or is not running.
 	})
 	xcheckf(err, "updating database")
 }
+
+func cmdFixmsgsize(c *cmd) {
+	c.params = "[account]"
+	c.help = `Ensure message sizes in the database matching the sum of the message prefix length and on-disk file size.
+
+Messages with an inconsistent size are also parsed again.
+
+If an inconsistency is found, you should probably also run "mox
+bumpuidvalidity" on the mailboxes or entire account to force IMAP clients to
+refetch messages.
+`
+	args := c.Parse()
+	if len(args) > 1 {
+		c.Usage()
+	}
+
+	mustLoadConfig()
+	var account string
+	if len(args) == 1 {
+		account = args[0]
+	}
+	ctlcmdFixmsgsize(xctl(), account)
+}
+
+func ctlcmdFixmsgsize(ctl *ctl, account string) {
+	ctl.xwrite("fixmsgsize")
+	ctl.xwrite(account)
+	ctl.xreadok()
+	ctl.xstreamto(os.Stdout)
+}
+
+func cmdReparse(c *cmd) {
+	c.params = "[account]"
+	c.help = `Parse all messages in the account or all accounts again
+
+Can be useful after upgrading mox with improved message parsing. Messages are
+parsed in batches, so other access to the mailboxes/messages are not blocked
+while reparsing all messages.
+`
+	args := c.Parse()
+	if len(args) > 1 {
+		c.Usage()
+	}
+
+	mustLoadConfig()
+	var account string
+	if len(args) == 1 {
+		account = args[0]
+	}
+	ctlcmdReparse(xctl(), account)
+}
+
+func ctlcmdReparse(ctl *ctl, account string) {
+	ctl.xwrite("reparse")
+	ctl.xwrite(account)
+	ctl.xreadok()
+	ctl.xstreamto(os.Stdout)
+}
+
+func cmdEnsureParsed(c *cmd) {
+	c.params = "account"
+	c.help = "Ensure messages in the database have a pre-parsed MIME form in the database."
+	var all bool
+	c.flag.BoolVar(&all, "all", false, "store new parsed message for all messages")
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
+	}
+
+	clog := mlog.New("ensureparsed")
+
+	mustLoadConfig()
+	a, err := store.OpenAccount(args[0])
+	xcheckf(err, "open account")
+	defer func() {
+		if err := a.Close(); err != nil {
+			log.Printf("closing account: %v", err)
+		}
+	}()
+
+	n := 0
+	err = a.DB.Write(context.Background(), func(tx *bstore.Tx) error {
+		q := bstore.QueryTx[store.Message](tx)
+		q.FilterEqual("Expunged", false)
+		q.FilterFn(func(m store.Message) bool {
+			return all || m.ParsedBuf == nil
+		})
+		l, err := q.List()
+		if err != nil {
+			return fmt.Errorf("list messages: %v", err)
+		}
+		for _, m := range l {
+			mr := a.MessageReader(m)
+			p, err := message.EnsurePart(clog, false, mr, m.Size)
+			if err != nil {
+				log.Printf("parsing message %d: %v (continuing)", m.ID, err)
+			}
+			m.ParsedBuf, err = json.Marshal(p)
+			if err != nil {
+				return fmt.Errorf("marshal parsed message: %v", err)
+			}
+			if err := tx.Update(&m); err != nil {
+				return fmt.Errorf("update message: %v", err)
+			}
+			n++
+		}
+		return nil
+	})
+	xcheckf(err, "update messages with parsed mime structure")
+	fmt.Printf("%d messages updated\n", n)
+}
+
+func cmdRecalculateMailboxCounts(c *cmd) {
+	c.params = "account"
+	c.help = `Recalculate message counts for all mailboxes in the account.
+
+When a message is added to/removed from a mailbox, or when message flags change,
+the total, unread, unseen and deleted messages are accounted, and the total size
+of the mailbox. In case of a bug in this accounting, the numbers could become
+incorrect. This command will find, fix and print them.
+`
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
+	}
+
+	mustLoadConfig()
+	ctlcmdRecalculateMailboxCounts(xctl(), args[0])
+}
+
+func ctlcmdRecalculateMailboxCounts(ctl *ctl, account string) {
+	ctl.xwrite("recalculatemailboxcounts")
+	ctl.xwrite(account)
+	ctl.xreadok()
+	ctl.xstreamto(os.Stdout)
+}
+
+func cmdMessageParse(c *cmd) {
+	c.params = "message.eml"
+	c.help = "Parse message, print JSON representation."
+
+	args := c.Parse()
+	if len(args) != 1 {
+		c.Usage()
+	}
+
+	clog := mlog.New("messageparse")
+
+	f, err := os.Open(args[0])
+	xcheckf(err, "open")
+	defer f.Close()
+
+	part, err := message.Parse(clog, false, f)
+	xcheckf(err, "parsing message")
+	err = part.Walk(clog, nil)
+	xcheckf(err, "parsing nested parts")
+	enc := json.NewEncoder(os.Stdout)
+	enc.SetIndent("", "\t")
+	err = enc.Encode(part)
+	xcheckf(err, "write")
+}
+
+func cmdOpenaccounts(c *cmd) {
+	c.unlisted = true
+	c.params = "datadir account ..."
+	c.help = `Open and close accounts, for triggering data upgrades, for tests.
+
+Opens database files directly, not going through a running mox instance.
+`
+
+	args := c.Parse()
+	if len(args) <= 1 {
+		c.Usage()
+	}
+
+	dataDir := filepath.Clean(args[0])
+	for _, accName := range args[1:] {
+		accDir := filepath.Join(dataDir, "accounts", accName)
+		log.Printf("opening account %s...", filepath.Join(accDir, accName))
+		a, err := store.OpenAccountDB(accDir, accName)
+		xcheckf(err, "open account %s", accName)
+		err = a.Close()
+		xcheckf(err, "close account %s", accName)
+	}
+}
