@@ -1,4 +1,4 @@
 package message
 
 // Timestamp as used in internet mail messages.
-const RFC5322Z = "02 Jan 2006 15:04 -0700"
+const RFC5322Z = "2 Jan 2006 15:04:05 -0700"
