@@ -272,7 +272,15 @@ func (c *Config) allowACMEHosts(checkACMEHosts bool) {
 			}
 		}
 
-		m.SetAllowedHostnames(dns.StrictResolver{Pkg: "autotls"}, hostnames, c.Static.Listeners["public"].IPs, checkACMEHosts)
+		public := c.Static.Listeners["public"]
+		ips := public.IPs
+		if len(public.NATIPs) > 0 {
+			ips = public.NATIPs
+		}
+		if public.IPsNATed {
+			ips = nil
+		}
+		m.SetAllowedHostnames(dns.StrictResolver{Pkg: "autotls"}, hostnames, ips, checkACMEHosts)
 	}
 }
 
@@ -629,6 +637,17 @@ func PrepareStaticConfig(ctx context.Context, configFile string, conf *Config, c
 			}
 			l.SMTP.DNSBLZones = append(l.SMTP.DNSBLZones, d)
 		}
+		if l.IPsNATed && len(l.NATIPs) > 0 {
+			addErrorf("listener %q has both IPsNATed and NATIPs (remove deprecated IPsNATed)", name)
+		}
+		for _, ipstr := range l.NATIPs {
+			ip := net.ParseIP(ipstr)
+			if ip == nil {
+				addErrorf("listener %q has invalid ip %q", name, ipstr)
+			} else if ip.IsUnspecified() || ip.IsLoopback() {
+				addErrorf("listener %q has NAT ip that is the unspecified or loopback address %s", name, ipstr)
+			}
+		}
 		checkPath := func(kind string, enabled bool, path string) {
 			if enabled && path != "" && !strings.HasPrefix(path, "/") {
 				addErrorf("listener %q has %s with path %q that must start with a slash", name, kind, path)
@@ -644,9 +663,33 @@ func PrepareStaticConfig(ctx context.Context, configFile string, conf *Config, c
 		c.SpecifiedSMTPListenIPs = nil
 	}
 
+	var zerouse config.SpecialUseMailboxes
+	if len(c.DefaultMailboxes) > 0 && (c.InitialMailboxes.SpecialUse != zerouse || len(c.InitialMailboxes.Regular) > 0) {
+		addErrorf("cannot have both DefaultMailboxes and InitialMailboxes")
+	}
+	// DefaultMailboxes is deprecated.
 	for _, mb := range c.DefaultMailboxes {
 		checkMailboxNormf(mb, "default mailbox")
 	}
+	checkSpecialUseMailbox := func(nameOpt string) {
+		if nameOpt != "" {
+			checkMailboxNormf(nameOpt, "special-use initial mailbox")
+			if strings.EqualFold(nameOpt, "inbox") {
+				addErrorf("initial mailbox cannot be set to Inbox (Inbox is always created)")
+			}
+		}
+	}
+	checkSpecialUseMailbox(c.InitialMailboxes.SpecialUse.Archive)
+	checkSpecialUseMailbox(c.InitialMailboxes.SpecialUse.Draft)
+	checkSpecialUseMailbox(c.InitialMailboxes.SpecialUse.Junk)
+	checkSpecialUseMailbox(c.InitialMailboxes.SpecialUse.Sent)
+	checkSpecialUseMailbox(c.InitialMailboxes.SpecialUse.Trash)
+	for _, name := range c.InitialMailboxes.Regular {
+		checkMailboxNormf(name, "regular initial mailbox")
+		if strings.EqualFold(name, "inbox") {
+			addErrorf("initial regular mailbox cannot be set to Inbox (Inbox is always created)")
+		}
+	}
 
 	checkTransportSMTP := func(name string, isTLS bool, t *config.TransportSMTP) {
 		var err error
@@ -1060,6 +1103,14 @@ func prepareDynamicConfig(ctx context.Context, dynamicPath string, static config
 					addErrorf("ruleset must have at least one rule")
 				}
 
+				if rs.IsForward && rs.ListAllowDomain != "" {
+					addErrorf("ruleset cannot have both IsForward and ListAllowDomain")
+				}
+				if rs.IsForward {
+					if rs.SMTPMailFromRegexp == "" || rs.VerifiedDomain == "" {
+						addErrorf("ruleset with IsForward must have both SMTPMailFromRegexp and VerifiedDomain too")
+					}
+				}
 				if rs.ListAllowDomain != "" {
 					d, err := dns.ParseDomain(rs.ListAllowDomain)
 					if err != nil {
@@ -1067,6 +1118,11 @@ func prepareDynamicConfig(ctx context.Context, dynamicPath string, static config
 					}
 					c.Accounts[accName].Destinations[addrName].Rulesets[i].ListAllowDNSDomain = d
 				}
+
+				checkMailboxNormf(rs.AcceptRejectsToMailbox, "account %q, destination %q, ruleset %d, rejects mailbox", accName, addrName, i+1)
+				if strings.EqualFold(rs.AcceptRejectsToMailbox, "inbox") {
+					addErrorf("account %q, destination %q, ruleset %d: AcceptRejectsToMailbox cannot be set to Inbox", accName, addrName, i+1)
+				}
 			}
 
 			// Catchall destination for domain.
