@@ -101,9 +101,10 @@ type Msg struct {
 	NextAttempt        time.Time           // For scheduling.
 	LastAttempt        *time.Time
 	LastError          string
-	Has8bit            bool  // Whether message contains bytes with high bit set, determines whether 8BITMIME SMTP extension is needed.
-	SMTPUTF8           bool  // Whether message requires use of SMTPUTF8.
-	Size               int64 // Full size of message, combined MsgPrefix with contents of message file.
+	Has8bit            bool   // Whether message contains bytes with high bit set, determines whether 8BITMIME SMTP extension is needed.
+	SMTPUTF8           bool   // Whether message requires use of SMTPUTF8.
+	Size               int64  // Full size of message, combined MsgPrefix with contents of message file.
+	MessageID          string // Used when composing a DSN, in its References header.
 	MsgPrefix          []byte
 
 	// If set, this message is a DSN and this is a version using utf-8, for the case
@@ -199,12 +200,32 @@ func Count(ctx context.Context) (int, error) {
 // this data is used as the message when delivering the DSN and the remote SMTP
 // server supports SMTPUTF8. If the remote SMTP server does not support SMTPUTF8,
 // the regular non-utf8 message is delivered.
-func Add(ctx context.Context, log *mlog.Log, senderAccount string, mailFrom, rcptTo smtp.Path, has8bit, smtputf8 bool, size int64, msgPrefix []byte, msgFile *os.File, dsnutf8Opt []byte, consumeFile bool) (int64, error) {
+func Add(ctx context.Context, log *mlog.Log, senderAccount string, mailFrom, rcptTo smtp.Path, has8bit, smtputf8 bool, size int64, messageID string, msgPrefix []byte, msgFile *os.File, dsnutf8Opt []byte, consumeFile bool) (int64, error) {
 	// todo: Add should accept multiple rcptTo if they are for the same domain. so we can queue them for delivery in one (or just a few) session(s), transferring the data only once. ../rfc/5321:3759
 
 	if Localserve {
-		// Safety measure, shouldn't happen.
-		return 0, fmt.Errorf("no queuing with localserve")
+		if senderAccount == "" {
+			return 0, fmt.Errorf("cannot queue with localserve without local account")
+		}
+		acc, err := store.OpenAccount(senderAccount)
+		if err != nil {
+			return 0, fmt.Errorf("opening sender account for immediate delivery with localserve: %v", err)
+		}
+		defer func() {
+			err := acc.Close()
+			log.Check(err, "closing account")
+		}()
+		m := store.Message{Size: size, MsgPrefix: msgPrefix}
+		conf, _ := acc.Conf()
+		dest := conf.Destinations[mailFrom.String()]
+		acc.WithWLock(func() {
+			err = acc.Deliver(log, dest, &m, msgFile, consumeFile)
+		})
+		if err != nil {
+			return 0, fmt.Errorf("delivering message: %v", err)
+		}
+		log.Debug("immediately delivered from queue to sender")
+		return 0, nil
 	}
 
 	tx, err := DB.Begin(ctx, true)
@@ -220,7 +241,7 @@ func Add(ctx context.Context, log *mlog.Log, senderAccount string, mailFrom, rcp
 	}()
 
 	now := time.Now()
-	qm := Msg{0, now, senderAccount, mailFrom.Localpart, mailFrom.IPDomain, rcptTo.Localpart, rcptTo.IPDomain, formatIPDomain(rcptTo.IPDomain), 0, nil, now, nil, "", has8bit, smtputf8, size, msgPrefix, dsnutf8Opt, ""}
+	qm := Msg{0, now, senderAccount, mailFrom.Localpart, mailFrom.IPDomain, rcptTo.Localpart, rcptTo.IPDomain, formatIPDomain(rcptTo.IPDomain), 0, nil, now, nil, "", has8bit, smtputf8, size, messageID, msgPrefix, dsnutf8Opt, ""}
 
 	if err := tx.Insert(&qm); err != nil {
 		return 0, err
@@ -240,14 +261,9 @@ func Add(ctx context.Context, log *mlog.Log, senderAccount string, mailFrom, rcp
 			// Could be due to cross-filesystem rename. Users shouldn't configure their systems that way.
 			return 0, fmt.Errorf("move message into queue dir: %w", err)
 		}
-	} else if err := os.Link(msgFile.Name(), dst); err != nil {
-		// Assume file system does not support hardlinks. Copy it instead.
-		if err := writeFile(dst, &moxio.AtReader{R: msgFile}); err != nil {
-			return 0, fmt.Errorf("copying message to new file: %s", err)
-		}
-	}
-
-	if err := moxio.SyncDir(dstDir); err != nil {
+	} else if err := moxio.LinkOrCopy(log, dst, msgFile.Name(), nil, true); err != nil {
+		return 0, fmt.Errorf("linking/copying message to new file: %s", err)
+	} else if err := moxio.SyncDir(dstDir); err != nil {
 		return 0, fmt.Errorf("sync directory: %v", err)
 	}
 
@@ -261,29 +277,6 @@ func Add(ctx context.Context, log *mlog.Log, senderAccount string, mailFrom, rcp
 	return qm.ID, nil
 }
 
-// write contents of r to new file dst, for delivering a message.
-func writeFile(dst string, r io.Reader) error {
-	df, err := os.OpenFile(dst, os.O_CREATE|os.O_EXCL|os.O_WRONLY, 0660)
-	if err != nil {
-		return fmt.Errorf("create: %w", err)
-	}
-	defer func() {
-		if df != nil {
-			err := df.Close()
-			xlog.Check(err, "closing file after failed write")
-		}
-	}()
-	if _, err := io.Copy(df, r); err != nil {
-		return fmt.Errorf("copy: %s", err)
-	} else if err := df.Sync(); err != nil {
-		return fmt.Errorf("sync: %s", err)
-	} else if err := df.Close(); err != nil {
-		return fmt.Errorf("close: %s", err)
-	}
-	df = nil
-	return nil
-}
-
 func formatIPDomain(d dns.IPDomain) string {
 	if len(d.IP) > 0 {
 		return "[" + d.IP.String() + "]"
