@@ -186,9 +186,9 @@ requested, other TLS certificates are requested on demand.
 			}
 		}
 	} else {
-		log.Print("starting as unprivileged user", mlog.Field("user", mox.Conf.Static.User), mlog.Field("uid", mox.Conf.Static.UID), mlog.Field("gid", mox.Conf.Static.GID), mlog.Field("pid", os.Getpid()))
 		mox.RestorePassedFiles()
 		mox.MustLoadConfig(true, checkACMEHosts)
+		log.Print("starting as unprivileged user", mlog.Field("user", mox.Conf.Static.User), mlog.Field("uid", mox.Conf.Static.UID), mlog.Field("gid", mox.Conf.Static.GID), mlog.Field("pid", os.Getpid()))
 	}
 
 	syscall.Umask(syscall.Umask(007) | 007)
@@ -301,8 +301,11 @@ requested, other TLS certificates are requested on demand.
 					log.Check(err, "closing temp changelog file")
 				}
 			}()
-			m := &store.Message{Received: time.Now(), Flags: store.Flags{Flagged: true}}
-			n, err := fmt.Fprintf(f, "Date: %s\r\nSubject: mox %s available\r\n\r\nHi!\r\n\r\nVersion %s of mox is available, this install is at %s.\r\n\r\nChanges:\r\n\r\n%s\r\n\r\nRemember to make a backup with \"mox backup\" before upgrading.\r\nPlease report any issues at https://github.com/mjl-/mox, thanks!\r\n\r\nCheers,\r\nmox\r\n", time.Now().Format(message.RFC5322Z), latest, latest, current, strings.ReplaceAll(cl, "\n", "\r\n"))
+			m := &store.Message{
+				Received: time.Now(),
+				Flags:    store.Flags{Flagged: true},
+			}
+			n, err := fmt.Fprintf(f, "Date: %s\r\nSubject: mox %s available\r\nContent-Type: text/plain; charset=utf-8\r\nContent-Transfer-Encoding: 8-bit\r\n\r\nHi!\r\n\r\nVersion %s of mox is available, this install is at %s.\r\n\r\nChanges:\r\n\r\n%s\r\n\r\nRemember to make a backup with \"mox backup\" before upgrading.\r\nPlease report any issues at https://github.com/mjl-/mox, thanks!\r\n\r\nCheers,\r\nmox\r\n", time.Now().Format(message.RFC5322Z), latest, latest, current, strings.ReplaceAll(cl, "\n", "\r\n"))
 			if err != nil {
 				log.Infox("writing temporary message file for changelog delivery", err)
 				return next
@@ -361,7 +364,7 @@ requested, other TLS certificates are requested on demand.
 		for _, e := range tmps {
 			if fi, err := e.Info(); err != nil {
 				log.Errorx("stat tmp file", err, mlog.Field("filename", e.Name()))
-			} else if now.Sub(fi.ModTime()) > 7*24*time.Hour {
+			} else if now.Sub(fi.ModTime()) > 7*24*time.Hour && !fi.IsDir() {
 				p := filepath.Join(tmpdir, e.Name())
 				if err := os.Remove(p); err != nil {
 					log.Errorx("removing stale temporary file", err, mlog.Field("path", p))
@@ -621,7 +624,8 @@ func start(mtastsdbRefresher, skipForkExec bool) error {
 	http.Serve()
 
 	go func() {
-		<-store.Switchboard()
+		store.Switchboard()
+		<-make(chan struct{})
 	}()
 	return nil
 }
