@@ -18,7 +18,7 @@ import (
 
 // rejectPresent returns whether the message is already present in the rejects mailbox.
 func rejectPresent(log *mlog.Log, acc *store.Account, rejectsMailbox string, m *store.Message, f *os.File) (present bool, msgID string, hash []byte, rerr error) {
-	if p, err := message.Parse(store.FileMsgReader(m.MsgPrefix, f)); err != nil {
+	if p, err := message.Parse(log, false, store.FileMsgReader(m.MsgPrefix, f)); err != nil {
 		log.Infox("parsing reject message for message-id", err)
 	} else if header, err := p.Header(); err != nil {
 		log.Infox("parsing reject message header for message-id", err)
@@ -54,6 +54,7 @@ func rejectPresent(log *mlog.Log, acc *store.Account, rejectsMailbox string, m *
 
 			q := bstore.QueryTx[store.Message](tx)
 			q.FilterNonzero(store.Message{MailboxID: mb.ID})
+			q.FilterEqual("Expunged", false)
 			q.FilterFn(func(m store.Message) bool {
 				return msgID != "" && m.MessageID == msgID || len(hash) > 0 && bytes.Equal(m.MessageHash, hash)
 			})
